/*************************************************
 * tclprof.c
 *
 * TCL profiler for use with AOLserver.
 * Copyright (C) 2002  Eric Lorenzo (eric@openforce.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *************************************************/

#define USE_TCL8X

#include <tcl.h>

/* Including this file is slightly evil, but we need the Interp struct */
#include <tclInt.h>

#include <ns.h>

#include <stdlib.h>

#include <unistd.h>

#include <string.h>

#include <limits.h>



/***********************************************************************
 Glossary Of Library Calls: A listing of Tcl and AOLserver library
 calls and data structures used in this program.

 TCL (Tcl_*):

   Interpreter data structure:
     (http://www.scriptics.com/man/tcl8.3/TclLib/Interp.htm)

     + Tcl_Interp: Tcl Interpreter

   Generating a TCL result from a C procedure implementing a TCL command:
     (http://www.scriptics.com/man/tcl8.3/TclLib/SetResult.htm)

     - Tcl_AppendResult(interp, string, string, ..., (char *)NULL);

     - Tcl_SetResult(interp, string, freeProc);

     - Tcl_SetObjResult(interp, string);

   Installing a trace listener:
     (http://www.scriptics.com/man/tcl8.3/TclLib/CrtTrace.htm)

     - Tcl_CreateTrace(interp, INT_MAX, &ProfileTrace, rt);
   
   Registering a C function as a TCL command:
     (http://www.scriptics.com/man/tcl8.3/TclLib/CrtCommand.htm)

     + Tcl_CmdProc: A pointer to a C function implementing a TCL Command

     - Tcl_CreateCommand(interp, cmdName, cmdProc, clientData, deleteProc);

   Hash Tables:
     (http://www.scriptics.com/man/tcl8.3/TclLib/CrtCommand.htm)

     + Tcl_HashTable: A TCL hash table

     + Tcl_HashEntry: An entry in a TCL hash table

     + Tcl_HashSearch: The state of a hash table walk

     - Tcl_InitHashTable(tablePtr, keyType);

     _ Tcl_DeleteHashTable(tablePtr);

     - Tcl_CreateHashEntry(tablePtr, key, newPtr);

     - Tcl_FindHashEntry(tablePtr, key);

     - Tcl_SetHashValue(entryPtr, value);

     - Tcl_GetHashValue(entryPtr);

     - Tcl_FirstHashEntry(treePtr, searchPtr);

     - Tcl_NextHashEntry(searchPtr);


 Tcl_Objs:

     - Tcl_NewObj();
       (http://www.scriptics.com/man/tcl8.3/TclLib/Object.htm)

     - Tcl_NewListObj(objc, objv);
     - Tcl_ListObjAppendElement(interp, listPtr, objPtr);
       (http://www.scriptics.com/man/tcl8.3/TclLib/ListObj.htm)

     - Tcl_NewStringObj(bytes, length);
     - Tcl_GetString(objPtr);
       (http://www.scriptics.com/man/tcl8.3/TclLib/StringObj.htm);

     - Tcl_NewIntObj(intValue);
       (http://www.scriptics.com/man/tcl8.3/TclLib/IntObj.htm);

     - Tcl_DoubleIntObj(doubleValue);
       (http://www.scriptics.com/man/tcl8.3/TclLib/DoubleObj.htm)

     
 AOLserver (Ns_*):

   System Time:

     + Ns_Time: Data structure storing a time value

     - Ns_GetTime(Ns_Time *);
       (http://www.aolserver.com/docs/devel/c/api/c-ch210.htm)


   Memory Allocation:

     - Ns_Malloc(size_t);
       (http://www.aolserver.com/docs/devel/c/api/c-ch248.htm)

     - Ns_Calloc(size_t, size_t);
       (http://www.aolserver.com/docs/devel/c/api/c-ch42.htm)

     - Ns_Realloc(void *, size_t);
       (http://www.aolserver.com/docs/devel/c/api/c-ch284.htm)

     - Ns_Free(void *);
       (http://www.aolserver.com/docs/devel/c/api/c-ch193.htm)


   Simple Mutexes:

     + Ns_Mutex:  A simple mutex

     - Ns_MutexInit(mutex);
       (http://www.aolserver.com/docs/devel/c/api/c-ch261.htm)

     - Ns_MutexLock(mutex);
       (http://www.aolserver.com/docs/devel/c/api/c-ch262.htm)

     - Ns_MutexUnlock(mutex);
       (http://www.aolserver.com/docs/devel/c/api/c-ch263.htm)


   Read/Write Locks:

     + Ns_RWLock: A read/write lock

     - Ns_RWLockInit(rwlock);
       (http://www.aolserver.com/docs/devel/c/api/c-ch305.htm)

     - Ns_RWLockRdLock(rwlock);
       (http://www.aolserver.com/docs/devel/c/api/c-ch306.htm)

     - Ns_RWLockWrLock(rwlock);
       (http://www.aolserver.com/docs/devel/c/api/c-ch308.htm)

     - Ns_RWLockUnlock(rwlock);
       (http://www.aolserver.com/docs/devel/c/api/c-ch307.htm)
     

    Thread-Local storage:

      + Ns_Tls:  A variable with thread-local values

      - Ns_TlsAlloc(tls);
        (http://www.aolserver.com/docs/devel/c/api/c-ch413.htm)

      - Ns_TlsGet(tls);
        (http://www.aolserver.com/docs/devel/c/api/c-ch414.htm)

      - Ns_TlsSet(tls);
        (http://www.aolserver.com/docs/devel/c/api/c-ch415.htm)


    Request Data:

      + Ns_Conn: The data AOLserver stores about a request connection
        (http://www.aolserver.com/docs/devel/c/api/c-ch3.htm#33642)

      - Ns_TclGetConn(NULL);
        (http://www.aolserver.com/docs/devel/c/api/c-ch387.htm)


    Logging:

      - Ns_Log(severity, format, ...);
        (http://www.aolserver.com/docs/devel/c/api/c-ch244.htm)


    Interpreter Configuration:

      - Ns_TclRegisterAtCreate(initProc, arg);
        (http://www.aolserver.com/docs/devel/c/api/c-ch398.htm)

      - Ns_TclRegisterAtDelete(deleteProc, arg);
        (UNDOCUMENTED)

      - Ns_TclInitInterps(server, initProc, arg);
        (http://www.aolserver.com/docs/devel/c/api/c-ch392.htm)


    Request Handling:

      - Ns_RegisterCleanup(cleanupProc, arg);
        (http://www.aolserver.com/docs/devel/c/api/c-ch291.htm)


    Scheduled Procs:

      - Ns_ScheduleProc(scheduleProc, arg, newThreadP, interval);
        (http://www.aolserver.com/docs/devel/c/api/c-ch310.htm)

 ***********************************************************************/



/* Required by AOLserver */
NS_EXPORT int Ns_ModuleVersion = 1;



/* Standard config-file section for profiler parameters */
#define TCL_TRACE_PARAMETER_SECTION "tclprof/parameters"

/* Names of TCL commands */
#define TCL_TRACE_CURRENT_REQUEST_CMD "tt_current_request_number"
#define TCL_TRACE_ENABLED_CMD "tt_trace_enabled_p"
#define TCL_TRACE_GET_REQUEST_LISTING_CMD "tt_get_request_list"
#define TCL_TRACE_GET_REQUEST_INFO_CMD "tt_get_request_info"
#define TCL_TRACE_GET_REQUEST_TRACE_CMD "tt_get_request_trace"
#define TCL_TRACE_GET_REQUEST_STATISTICS_CMD "tt_get_request_statistics"

/* Handy utility proc used in a few places */
double convertTime(Ns_Time t) {
  double result = t.usec;
  result /= 1000000.0;
  result += t.sec;
  return result;
}






/************************************************************************
 MemoryPool -- used to manage memory allocation during a request

 We have a situation where we have to store a lot of short strings
 that all have the same lifetime: Namely, that of the RequestTrace to
 which they are attached.  The MemoryPool is designed to permit quick
 storage of these strings, avoiding calls to global malloc most of the
 time, and to make it easy to free the whole thing when the associated
 RequestTrace gets destroyed.

 The MemoryPool works by keeping a set of large blocks of memory, only
 one of which is 'active' at any given time.  When a request is made
 to store a string, we first check if the string will fit into the
 currently active block.  If not, a new block is allocated.

 Each block is the same size, and new MemoryPools are created with a
 single block.  Ideally, the block size should be large enough that we
 never have to allocate a new block during the course of a request.
 ************************************************************************/

#define MEMORY_POOL_BLOCK_SIZE_KEY "RequestStringBlockSize"
#define MEMORY_POOL_BLOCK_SIZE_DEFAULT 1048576
int g_memoryPoolBlockSize;

typedef struct _memoryPool {
  int totalAlloc;
  int currentBlock;
  int currentBlockUsed;
  int maxBlocks;
  void **blocks;
} MemoryPool;


/*******************************************
 * Allocate a new MemoryPool
 *******************************************/
MemoryPool *MpCreate() {
  MemoryPool *mp = (MemoryPool *)Ns_Malloc(sizeof(MemoryPool));  /* Freed in MpDestroy */
  mp->totalAlloc = 0;
  mp->maxBlocks = 4;
  mp->blocks = (void **)Ns_Malloc(mp->maxBlocks * sizeof(void *));  /* Freed in MpDestroy */
  mp->blocks[0] = Ns_Malloc(g_memoryPoolBlockSize);  /* Freed in MpDestroy */
  mp->currentBlock = 0;
  mp->currentBlockUsed = 0;
  return mp;
}



/*******************************************
 * Free up a MemoryPool
 *******************************************/
void MpDestroy(MemoryPool *mp) {
  int i;
  for (i = 0; i <= mp->currentBlock; i++) {
    Ns_Free(mp->blocks[i]);
  }
  Ns_Free(mp->blocks);
  Ns_Free(mp);
}


/******************************************
 * Allocate space from a MemoryPool
 ******************************************/
void *MpAlloc(MemoryPool *mp, size_t len) {
  void *dest;

  mp->totalAlloc += len;
  
  /* Check whether or not we need to allocate a new block.
     Hopefully this code is never actually needed. */
  if (len > (g_memoryPoolBlockSize - mp->currentBlockUsed)) {

    mp->currentBlock++;

    /* See if we need to expand the block pointer array */
    if (mp->currentBlock >= mp->maxBlocks) {
      mp->maxBlocks <<= 1;
      mp->blocks = (void **)Ns_Realloc(mp->blocks,
                                       mp->maxBlocks * sizeof(void *));
    }

    /* Special handling for allocations that exceed the standard block size
       Hopefully this code is never actually needed. */
    if (len > g_memoryPoolBlockSize) {
      /* Move the current active block into the new slot */
      mp->blocks[mp->currentBlock] = mp->blocks[mp->currentBlock - 1];
      /* Put a block big enough to accomodate the allocation request 
         into the previously active slot */
      mp->blocks[mp->currentBlock-1] = Ns_Malloc(len);  /* Freed in MpDestroy */
      /* return the newly allocated block */
      return mp->blocks[mp->currentBlock-1];
    }

    /* 'normal' case -- allocate a new block in the new slot */
    mp->blocks[mp->currentBlock] =
      Ns_Malloc(g_memoryPoolBlockSize); /* Freed in MpDestroy */
    mp->currentBlockUsed = 0;
  }

  dest = mp->blocks[mp->currentBlock] + mp->currentBlockUsed;
  mp->currentBlockUsed += len;
  
  return dest;
}


/*******************************************
 * Store a chunk of data in a MemoryPool
 ******************************************/
void *MpStore(MemoryPool *mp, const void *src, size_t srclen) {
  return memcpy(MpAlloc(mp, srclen), src, srclen);
}


/*******************************************
 * Store a string in a MemoryPool
 ******************************************/
char *MpStoreString(MemoryPool *mp, const char *src) {
  return (char *)MpStore(mp, src, strlen(src) + 1);
}


/*******************************************
 * Store an array of strings in a MemoryPool
 *
 * Note that this duplicates not just the
 * array of pointers, but also the strings
 * themselves.
 ******************************************/
char **MpStoreStringArray(MemoryPool *mp, int count, const char **data) {
  int i;
  char **dest = (char **)MpAlloc(mp, count * sizeof(char *));

  for (i = 0; i < count; i++) {
    dest[i] = MpStoreString(mp, data[i]);
  }

  return dest;
}


void _initializeMemoryPool() {
  if (Ns_ConfigGetInt(TCL_TRACE_PARAMETER_SECTION,
                      MEMORY_POOL_BLOCK_SIZE_KEY,
                      &g_memoryPoolBlockSize) != NS_TRUE) {
    g_memoryPoolBlockSize = MEMORY_POOL_BLOCK_SIZE_DEFAULT;
  }
}





/************************************************************************
 Basic data structures for recording a trace

 StatementTrace stores the data associated with a single TCL proc
 invocation.  

 RequestTrace stores the data logged for a single request.  This
 consists of an ID number that is unique within the AOLserver
 instance, the HTTP request line (e.g. "GET /foo/bar HTTP/1.0"), and
 a sequence of TCL statements that have been logged for the request.

 RequestTraces come in two flavors: Request thread RTs are used while
 actually logging trace data.  Long-term RTs are used to store request
 data in an internal list for a while so that it can be retrieved and
 presented to users.
 ************************************************************************/

typedef struct _statementTrace {
  int level;
  union {
    Ns_Time tv;
    double d;
  } when;
  char *cmd;
} StatementTrace;


typedef struct _requestTrace {
  int requestNumber;
  char *requestLine;
  int stmtCount;
  int stmtCapacity;
  StatementTrace *statements;
  MemoryPool *strings;
} RequestTrace;



/*******************************************
 * Construct a new RequestTrace.
 *
 * If the fullInitP flag is not set, this will only initialize the
 * simple fields of the RequestTrace -- the statements pointer
 * and strings MemoryPool pointer will be set to NULL.
 *
 * This flag exists so that we can use this function to create
 * long-term RequestTraces, which use a different allocation scheme
 * than request thread RTs for their statements vector and inherit
 * their strings MemoryPool from a request thread RequestTrace.
 *******************************************/
RequestTrace *RtCreate(int initCapacity, int fullInitP) {
  /* This is destroyed when the RequestTrace is destroyed */
  RequestTrace *rt = (RequestTrace *)Ns_Malloc(sizeof(RequestTrace)); /* Freed in RtDestroy */

  rt->requestNumber = 0;
  rt->requestLine = NULL;

  rt->stmtCount = 0;

  rt->stmtCapacity = initCapacity;
  rt->statements =
    (StatementTrace *)Ns_Calloc(rt->stmtCapacity, sizeof(StatementTrace)); /* Freed in RtDestroy */

  if (fullInitP) {
    rt->strings = MpCreate();
  } else {
    rt->strings = NULL;
  }

  return rt;
}


/*******************************************
 * Destroy a RequestTrace
 *******************************************/
void RtDestroy(RequestTrace *rt) {
  if (rt->strings != NULL) {
    MpDestroy(rt->strings);
  }
  if (rt->statements != NULL) {
    Ns_Free(rt->statements);
  }
  Ns_Free(rt);
}












/************************************************************************
 The Core

 Procedures to install/deinstall the trace, and the trace itself.
 Also provides TCL command for access to current request number, which
 may later be used for integration with ACS developer support.
 ************************************************************************/

#define TRACE_INIT_ENABLED_KEY "Enabled"
#define TRACE_INIT_ENABLED_DEFAULT 0
int g_traceEnabledP;


#define TRACE_INIT_STMT_CAPACITY_KEY "TraceInitStatementCapacity"
#define TRACE_INIT_STMT_CAPACITY_DEFAULT 65536
int g_traceInitStmtCapacity;


/* This thread-local stores a pointer to the thread's RequestTrace structure */
Ns_Tls g_threadRequestTracePointer;

/* This global is used to assign numbers to requests that are unique
   within a server. */
Ns_Mutex g_requestNumberingMux;
int g_currentRequestNumber = 0;


/*******************************************
 * Log data about a TCL command invocation.
 *
 * This procedure is installed as a TCL interpreter trace, via the
 * Tcl_CreateTrace function.
 *******************************************/
void ProfileTrace(ClientData clientData,
                  Tcl_Interp *interp,
                  int level,
                  char *command, 
                  Tcl_CmdProc *cmdProc,
                  ClientData cmdClientData,
                  int argc,
                  char *argv[]) {
  int i;
  RequestTrace *rt;
  StatementTrace *st;
  Ns_Conn *conn;

  rt = (RequestTrace *)clientData;

  /* Special processing when nothing has yet been logged */
  if (rt->stmtCount == 0) {
    /* If there's no conn, then this isn't a connection request, and we don't trace */
    conn = Ns_TclGetConn(NULL);
    if (conn == NULL) {
      return;
    }
    /* We only start tracing at level 1 -- if we get a statement at a
       level higher than one in this case, then it's because we're
       inside a call to "_ns_cleanupinterp" that we're ignoring. */
    if (level > 1) {
      return;
    }
    /* We don't want to trace _ns_cleanupinterp -- FIXME: Need cleaner solution for this */
    if (strcmp(argv[0], "_ns_cleanupinterp") == 0) {
      return;
    }

    Ns_MutexLock(&g_requestNumberingMux);
    rt->requestNumber = ++g_currentRequestNumber;
    Ns_MutexUnlock(&g_requestNumberingMux);
    rt->requestLine = MpStoreString(rt->strings, conn->request->line);
  }

  /* Expand the statement buffer, if necessary -- this is bad */
  while (rt->stmtCount >= rt->stmtCapacity) {
    rt->stmtCapacity <<= 1;
    Ns_Log(Notice, "TCLPROF: Expanding statement buffer to %d", rt->stmtCapacity);
    rt->statements =
      (StatementTrace *)Ns_Realloc(rt->statements,
                                   rt->stmtCapacity * sizeof(StatementTrace));
  }

  /* Populate the next available statement record */
  st = &(rt->statements[rt->stmtCount++]);
  st->level = level;
  Ns_GetTime(&(st->when.tv));
  st->cmd = MpStoreString(rt->strings, argv[0]);
}


/* TCL Command to fetch the current request number
   Returns the empty string if there's no active request */
int GetCurrentRequestNumberCmd(ClientData dummy,
                               Tcl_Interp *interp,
                               int argc,
                               Tcl_Obj *CONST argv[]) {
  RequestTrace *rt;
  char result[30];

  if (argc > 1) {
    Tcl_AppendResult(interp, "\"", 
                     Tcl_GetString(argv[0]), "\" takes no arguments.", NULL);
    return TCL_ERROR;
  }

  if (!g_traceEnabledP) {
    Tcl_SetResult(interp,
                  "Current trace request number not available: Trace disabled",
                  TCL_STATIC);
    return TCL_ERROR;
  }

  rt = (RequestTrace *)Ns_TlsGet(&g_threadRequestTracePointer);

  if ((rt == NULL) || (rt->requestNumber == 0)) {
    Tcl_SetResult(interp, "", TCL_STATIC);
    return TCL_OK;
  }
  
  sprintf(result, "%d", rt->requestNumber);
  Tcl_SetResult(interp, result, TCL_VOLATILE);
  return TCL_OK;    
}



/* TCL Command to determine whether or not tracing is enabled */
int GetTraceEnabledCmd(ClientData dummy,
                       Tcl_Interp *interp,
                       int argc,
                       Tcl_Obj *CONST argv[]) {
  if (argc > 1) {
    Tcl_AppendResult(interp,
                     "\"", Tcl_GetString(argv[0]),
                     "\" takes no arguments.", NULL);
    return TCL_ERROR;
  }

  if (g_traceEnabledP) {
    Tcl_SetResult(interp, "1", TCL_STATIC);
  } else {
    Tcl_SetResult(interp, "0", TCL_STATIC);
  }

  return TCL_OK;
}



/* The next three functions replace the standard implementations of
   the TCL 'for', 'if', and 'while' flow-control builtins.  The reason
   for this is that these three builtins invoke Tcl_ExprBooleanObj for
   their test expressions.  If the expression passed to this function
   contains a bracketed procedure call, that procedure call will
   appear to occur at the same stack depth as the builtin.  This
   confuses the profiler into thinking that the builtin has exited,
   and that children of the builtin (executed from the body scripts)
   are actually children of the proc invoked in the test expression.

   The code for these three replacement implementations is essentially
   identical to the code for the standard Tcl implementations, from
   which it was copied.  The only change made is that the numLevels
   field of the interpreter in which they are invoked is incremented
   before and decremented after their calls to Tcl_ExprBooleanObj.
   This causes procs invoked in test expressions passed to the
   flow-control builtins to appear to be children of the builtin,
   instead of siblings.

   Note that this requires touching with TCL's internal Interp data
   structure, which is slightly evil, but I don't know any other
   solution.
*/

int
Modified_ForObjCmd(dummy, interp, objc, objv)
    ClientData dummy;                   /* Not used. */
    Tcl_Interp *interp;                 /* Current interpreter. */
    int objc;                           /* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument objects. */
{
    Interp *iPtr = (Interp *) interp;
    int result, value;

    if (objc != 5) {
        Tcl_WrongNumArgs(interp, 1, objv, "start test next command");
        return TCL_ERROR;
    }

    result = Tcl_EvalObjEx(interp, objv[1], 0);
    if (result != TCL_OK) {
        if (result == TCL_ERROR) {
            Tcl_AddErrorInfo(interp, "\n    (\"for\" initial command)");
        }
        return result;
    }
    while (1) {
	/*
	 * We need to reset the result before passing it off to
	 * Tcl_ExprBooleanObj.  Otherwise, any error message will be appended
	 * to the result of the last evaluation.
	 */

	Tcl_ResetResult(interp);
        iPtr->numLevels++;
        result = Tcl_ExprBooleanObj(interp, objv[2], &value);
        iPtr->numLevels--;
        if (result != TCL_OK) {
            return result;
        }
        if (!value) {
            break;
        }
        result = Tcl_EvalObjEx(interp, objv[4], 0);
        if ((result != TCL_OK) && (result != TCL_CONTINUE)) {
            if (result == TCL_ERROR) {
                char msg[32 + TCL_INTEGER_SPACE];

                sprintf(msg, "\n    (\"for\" body line %d)",interp->errorLine);
                Tcl_AddErrorInfo(interp, msg);
            }
            break;
        }
        result = Tcl_EvalObjEx(interp, objv[3], 0);
	if (result == TCL_BREAK) {
            break;
        } else if (result != TCL_OK) {
            if (result == TCL_ERROR) {
                Tcl_AddErrorInfo(interp, "\n    (\"for\" loop-end command)");
            }
            return result;
        }
    }
    if (result == TCL_BREAK) {
        result = TCL_OK;
    }
    if (result == TCL_OK) {
        Tcl_ResetResult(interp);
    }
    return result;
}



int
Modified_IfObjCmd(dummy, interp, objc, objv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    Tcl_Obj *CONST objv[];		/* Argument objects. */
{
    Interp *iPtr = (Interp *) interp;
    int thenScriptIndex = 0;	/* then script to be evaled after syntax check */
    int i, result, value;
    char *clause;
    i = 1;
    while (1) {
	/*
	 * At this point in the loop, objv and objc refer to an expression
	 * to test, either for the main expression or an expression
	 * following an "elseif".  The arguments after the expression must
	 * be "then" (optional) and a script to execute if the expression is
	 * true.
	 */

	if (i >= objc) {
	    clause = Tcl_GetString(objv[i-1]);
	    Tcl_AppendResult(interp, "wrong # args: no expression after \"",
		    clause, "\" argument", (char *) NULL);
	    return TCL_ERROR;
	}
	if (!thenScriptIndex) {
            /* Make it look like this Expr is happening one step up the stack */
            iPtr->numLevels++;
	    result = Tcl_ExprBooleanObj(interp, objv[i], &value);
            iPtr->numLevels--;
	    if (result != TCL_OK) {
		return result;
	    }
	}
	i++;
	if (i >= objc) {
	    missingScript:
	    clause = Tcl_GetString(objv[i-1]);
	    Tcl_AppendResult(interp, "wrong # args: no script following \"",
		    clause, "\" argument", (char *) NULL);
	    return TCL_ERROR;
	}
	clause = Tcl_GetString(objv[i]);
	if ((i < objc) && (strcmp(clause, "then") == 0)) {
	    i++;
	}
	if (i >= objc) {
	    goto missingScript;
	}
	if (value) {
	    thenScriptIndex = i;
	    value = 0;
	}
	
	/*
	 * The expression evaluated to false.  Skip the command, then
	 * see if there is an "else" or "elseif" clause.
	 */

	i++;
	if (i >= objc) {
	    if (thenScriptIndex) {
		return Tcl_EvalObjEx(interp, objv[thenScriptIndex], 0);
	    }
	    return TCL_OK;
	}
	clause = Tcl_GetString(objv[i]);
	if ((clause[0] == 'e') && (strcmp(clause, "elseif") == 0)) {
	    i++;
	    continue;
	}
	break;
    }

    /*
     * Couldn't find a "then" or "elseif" clause to execute.  Check now
     * for an "else" clause.  We know that there's at least one more
     * argument when we get here.
     */

    if (strcmp(clause, "else") == 0) {
	i++;
	if (i >= objc) {
	    Tcl_AppendResult(interp,
		    "wrong # args: no script following \"else\" argument",
		    (char *) NULL);
	    return TCL_ERROR;
	}
    }
    if (i < objc - 1) {
	Tcl_AppendResult(interp,
		"wrong # args: extra words after \"else\" clause in \"if\" command",
		(char *) NULL);
	return TCL_ERROR;
    }
    if (thenScriptIndex) {
	return Tcl_EvalObjEx(interp, objv[thenScriptIndex], 0);
    }
    return Tcl_EvalObjEx(interp, objv[i], 0);
}




int
Modified_WhileObjCmd(dummy, interp, objc, objv)
    ClientData dummy;                   /* Not used. */
    Tcl_Interp *interp;                 /* Current interpreter. */
    int objc;                           /* Number of arguments. */
    Tcl_Obj *CONST objv[];       	/* Argument objects. */
{
    Interp *iPtr = (Interp *) interp;
    int result, value;

    if (objc != 3) {
	Tcl_WrongNumArgs(interp, 1, objv, "test command");
        return TCL_ERROR;
    }

    while (1) {
        iPtr->numLevels++;
        result = Tcl_ExprBooleanObj(interp, objv[1], &value);
        iPtr->numLevels--;
        if (result != TCL_OK) {
            return result;
        }
        if (!value) {
            break;
        }
        result = Tcl_EvalObjEx(interp, objv[2], 0);
        if ((result != TCL_OK) && (result != TCL_CONTINUE)) {
            if (result == TCL_ERROR) {
                char msg[32 + TCL_INTEGER_SPACE];

                sprintf(msg, "\n    (\"while\" body line %d)",
                        interp->errorLine);
                Tcl_AddErrorInfo(interp, msg);
            }
            break;
        }
    }
    if (result == TCL_BREAK) {
        result = TCL_OK;
    }
    if (result == TCL_OK) {
        Tcl_ResetResult(interp);
    }
    return result;
}






/*******************************************
 * InitializeInterpreter
 *
 * This procedure, which will be registered to execute against each
 * child TCL interpreter as it is created from the parent, takes care
 * of two things: It creates a thread-local RequestTrace structure
 * that will be used to log TCL commands executed by the specified
 * interpreter, and it installs the trace procedure that does the
 * actual logging.
 *******************************************/
int InitializeInterpreter(Tcl_Interp *interp, void *ignored) {
  RequestTrace *rt;

  Ns_Log(Notice, "TCLPROF: Initializing interpreter %x", (unsigned)interp);

  /* Set up storage for the thread-local RequestTrace pointer */
  rt = RtCreate(g_traceInitStmtCapacity, 1);
  Ns_TlsSet(&g_threadRequestTracePointer, (void *)rt);

  /* Replace the flow-control builtins that use Tcl_ExprBooleanObj */
  Tcl_CreateObjCommand(interp,
                       "for",
                       Modified_ForObjCmd,
                       NULL,
                       NULL);

  Tcl_CreateObjCommand(interp,
                       "if",
                       Modified_IfObjCmd,
                       NULL,
                       NULL);

  Tcl_CreateObjCommand(interp,
                       "while",
                       Modified_WhileObjCmd,
                       NULL,
                       NULL);


  /* Register the trace proc.  We pass the thread-local RequestTrace
     pointer pointer as clientData to the trace proc, so that the
     trace proc doesn't have to mess with the Ns_Tls. */
  Tcl_CreateTrace(interp, INT_MAX, &ProfileTrace, rt);

  return NS_OK;
}


/*******************************************
 * DeleteInterpreter
 *
 * This procedure will be registered to execute against each TCL
 * interpreter as it is destroyed.  Its only job is to release the
 * RequestTrace structure associated with the interpreter.
 *******************************************/
int DeleteInterpreter(Tcl_Interp *interp, void *ctx) {
  RequestTrace *rt;

  rt = (RequestTrace *)Ns_TlsGet(&g_threadRequestTracePointer);
  if (rt != NULL) {
    RtDestroy(rt);
  }

  return NS_OK;
}


/* This callback is used to create the tt_current_request_number
   command in the parent interpreter and tt_trace_enabled commands */
int _parentInitTrace(Tcl_Interp *interp, void *ctx) {
  Tcl_CreateObjCommand(interp,
                       TCL_TRACE_CURRENT_REQUEST_CMD,
                       GetCurrentRequestNumberCmd,
                       NULL,
                       NULL);

  Tcl_CreateObjCommand(interp,
                       TCL_TRACE_ENABLED_CMD,
                       GetTraceEnabledCmd,
                       NULL,
                       NULL);

  return NS_OK;
}


/* Initialization */
void _initializeTrace(char *hServer) {
  int enabled_p;

  if (Ns_ConfigGetBool(TCL_TRACE_PARAMETER_SECTION,
                       TRACE_INIT_ENABLED_KEY,
                       &g_traceEnabledP) != NS_TRUE) {
    g_traceEnabledP = TRACE_INIT_ENABLED_DEFAULT;
  }

  if (g_traceEnabledP) {
    if (Ns_ConfigGetInt(TCL_TRACE_PARAMETER_SECTION,
                        TRACE_INIT_STMT_CAPACITY_KEY,
                        &g_traceInitStmtCapacity) != NS_TRUE) {
      g_traceInitStmtCapacity = TRACE_INIT_STMT_CAPACITY_DEFAULT;
    }


    Ns_TlsAlloc(&g_threadRequestTracePointer, NULL);
    Ns_MutexInit(&g_requestNumberingMux);
    Ns_TclRegisterAtCreate(InitializeInterpreter, NULL);
    Ns_TclRegisterAtDelete(DeleteInterpreter, NULL);
    Ns_TclInitInterps(hServer, _parentInitTrace, NULL);
  }
}













/************************************************************************
 * Longer-term storage of request traces.
 *
 * The ProfileTrace section above records traces, but it has only a weak
 * understanding of what a 'request' is.  Specifically, it can only tell
 * when a request has begun, not when one has ended.  The procedures in
 * this section record when a request ends, copy the trace data for the
 * request into long-term storage, and reset the trace process so that
 * it's ready to start a new request.
 ************************************************************************/


#define STORED_TRACE_SWEEP_PERIOD_KEY "StoredTraceSweepPeriod"
#define STORED_TRACE_SWEEP_PERIOD_DEFAULT 60
int g_storedTraceSweepPeriod;

#define STORED_TRACE_MAX_AGE_KEY "StoredTraceMaxAge"
#define STORED_TRACE_MAX_AGE_DEFAULT 600
int g_storedTraceMaxAge;

/* This struct  just adds  an access time  field to  the RequestTrace.
   The lastAccess field is set to the current time when the request is
   first stored, and refreshed any time the RT is retrieved. */
typedef struct _requestRecord {
  Ns_Time lastAccess;
  RequestTrace *trace;
} RequestRecord;

/* Globals for long-term request storage */
Ns_RWLock g_storedRequestLock;
Tcl_HashTable g_storedRequests;


/*******************************************
 * stroreRequest
 *
 * Put a RequestTrace into long-term storage
 *******************************************/
void storeRequest(RequestTrace *rt) {
  int isNew;
  Tcl_HashEntry *entry;
  RequestRecord
    *old_rr, 
    *rr = (RequestRecord *)Ns_Malloc(sizeof(RequestRecord));  /* Freed in sweepRequests */

  Ns_GetTime(&(rr->lastAccess));
  rr->trace = rt;

  Ns_RWLockWrLock(&g_storedRequestLock);

  entry = Tcl_CreateHashEntry(&g_storedRequests,
                              (char *)rt->requestNumber,
                              &isNew);
  if (!isNew) {
    old_rr = (RequestRecord *)Tcl_GetHashValue(entry);
    RtDestroy(old_rr->trace);
    Ns_Free(old_rr);
  }

  Tcl_SetHashValue(entry, rr);

  Ns_RWLockUnlock(&g_storedRequestLock);
}


/* Retrieve a RequestTrace from long-term storage, by number */
RequestTrace *retrieveRequest(int requestNumber) {
  RequestRecord *rr;
  Tcl_HashEntry *entry;

  Ns_RWLockRdLock(&g_storedRequestLock);

  entry = Tcl_FindHashEntry(&g_storedRequests, (char *)requestNumber);
  if (entry == NULL) {
    Ns_RWLockUnlock(&g_storedRequestLock);
    return NULL;
  }

  rr = Tcl_GetHashValue(entry);

  Ns_GetTime(&(rr->lastAccess));

  Ns_RWLockUnlock(&g_storedRequestLock);

  return rr->trace;
}


/* Sweeper to clean up old request traces, so that we don't end up 
   with a memory leak. */
void sweepRequests(void *ignored) {
  Tcl_HashEntry *entry;
  Tcl_HashSearch search;
  RequestRecord *rr;
  Ns_Time nowTv;
  double now;

  Ns_Log(Notice, "TCLPROF: Starting old request sweep");

  /* Get the current time as a double */
  Ns_GetTime(&nowTv);
  now = convertTime(nowTv);

  Ns_RWLockWrLock(&g_storedRequestLock);

  for (entry = Tcl_FirstHashEntry(&g_storedRequests, &search);
       entry != NULL;
       entry = Tcl_NextHashEntry(&search)) {

    rr = Tcl_GetHashValue(entry);

    if ((now - convertTime(rr->lastAccess)) > g_storedTraceMaxAge) {

      RtDestroy(rr->trace);
      Ns_Free(rr);
      /* TCL docs say that deleting while searching is dangerous, but
         reading the hash table code convinces me it's safe to delete
         an entry that has already been returned by NextHashEntry. */
      Tcl_DeleteHashEntry(entry);

    }

  }

  Ns_RWLockUnlock(&g_storedRequestLock);
}



/* TCL Command to retrieve basic info on current stored requests */
int GetRequestListCmd(ClientData dummy,
                      Tcl_Interp *interp,
                      int argc,
                      Tcl_Obj *CONST argv[]) {
  Tcl_HashSearch search;
  Tcl_HashEntry *entry;
  RequestRecord *rr;
  int isFirstLine = 1;
  Tcl_Obj *result, *resultEntry[3];

  if (argc > 1) {
    Tcl_AppendResult(interp,
                     "\"", Tcl_GetString(argv[0]),
                     "\" takes no arguments.", NULL);
    return TCL_ERROR;
  }
  
  Ns_RWLockRdLock(&g_storedRequestLock);

  result = Tcl_NewObj();
  for (entry = Tcl_FirstHashEntry(&g_storedRequests, &search);
       entry != NULL;
       entry = Tcl_NextHashEntry(&search)) {

    rr = Tcl_GetHashValue(entry);

    resultEntry[0] = Tcl_NewIntObj(rr->trace->requestNumber);
    resultEntry[1] = Tcl_NewStringObj(rr->trace->requestLine, -1);
    resultEntry[2] = Tcl_NewIntObj((int)rr->trace->statements[0].when.d);

    Tcl_ListObjAppendElement(interp,
                             result,
                             Tcl_NewListObj(3, resultEntry));
  }

  Ns_RWLockUnlock(&g_storedRequestLock);

  Tcl_SetObjResult(interp, result);

  return TCL_OK;
}





/* TCL Command to retrieve 'header' info on a specific request */
int GetRequestInfoCmd(ClientData dummy,
                      Tcl_Interp *interp,
                      int argc,
                      Tcl_Obj *CONST argv[]) {
  RequestTrace *rt;
  int requestNumber;
  char stringBuffer[1000];
  Tcl_Obj *resultEntry[6];
  double startTime, endTime;

  if (argc != 2) {
    Tcl_AppendResult(interp,
                     "wrong # args: should be: \"",
                     Tcl_GetString(argv[0]), " requestNumber\"", NULL);
    return TCL_ERROR;
  }

  if (Tcl_GetIntFromObj(interp, argv[1], &requestNumber) != TCL_OK) {
    Tcl_SetResult(interp, "parameter must be an integer", TCL_STATIC);
    return TCL_ERROR;
  }

  rt = retrieveRequest(requestNumber);

  startTime = rt->statements[0].when.d;
  endTime = rt->statements[rt->stmtCount].when.d;

  resultEntry[0] = Tcl_NewStringObj(rt->requestLine, -1);
  resultEntry[1] = Tcl_NewDoubleObj(startTime);
  resultEntry[2] = Tcl_NewDoubleObj(endTime);
  resultEntry[3] = Tcl_NewDoubleObj(endTime - startTime);
  resultEntry[4] = Tcl_NewIntObj(rt->stmtCount);
  resultEntry[5] = Tcl_NewIntObj(rt->strings->totalAlloc);

  Tcl_SetObjResult(interp, Tcl_NewListObj(6, resultEntry));

  return TCL_OK;
}



/*******************************************
 * CleanupConnection
 *
 * This cleanup handler is responsible for actually marking
 * the 'end' of a request.  It copies the current thread's
 * RequestTrace into long-term storage, then re-initializes
 * the thread's RequestTrace so that the next invocation
 * of ProfileTrace will initiate a new request.
 *******************************************/
static void CleanupConnection(void *context, Ns_Conn *conn) {
  RequestTrace *rt, *newRt;
  StatementTrace *st;
  Ns_Time finalTime;
  int i;

  Ns_GetTime(&finalTime);

  rt = (RequestTrace *)Ns_TlsGet(&g_threadRequestTracePointer);
  if (rt == NULL) {
    return;
  }

  /* If rt is uninitialized, ignore this request */
  if (rt->requestNumber == 0) {
    return;
  }

  Ns_Log(Notice,
         "TCLPROF: Storing request data for request #%d (String alloc: %d)", 
         rt->requestNumber, rt->strings->totalAlloc);

  /* You may wonder why we choose to copy the data from the existing
     RequestTrace into a new one that will be put into long-term
     storage, rather than stuffing the existing RequestTrace into
     long-term storage and creating a new one for future logging by
     this thread.  The main reason is that creating a new one for
     long-term storage saves us a little memory, by allowing us to
     only allocate enough to store the exact number of statements that
     were logged. */

  /* Allocate the long-term RequestTrace and copy its simple fields from rt */
  newRt = RtCreate(rt->stmtCount + 2, 0);
  newRt->requestNumber = rt->requestNumber;
  newRt->requestLine = rt->requestLine;
  newRt->stmtCount = rt->stmtCount + 1;
  newRt->strings = rt->strings;

  /* Add fake 'toplevel' command to beginning of long-term trace, at
     level 0 This simplifies traversal later. */
  st = &(newRt->statements[0]);
  st->level = 0;
  st->when.d = convertTime(rt->statements[0].when.tv);
  st->cmd = "REQUEST_TOPLEVEL";

  /* Copy over real statements, converting times to doubles as we go */
  for (i = 0; i < rt->stmtCount; i++) {
    newRt->statements[i+1] = rt->statements[i];
    newRt->statements[i+1].when.d = convertTime(rt->statements[i].when.tv);
  }

  /* Add a 0-level sentinel to the end. */
  st = &(newRt->statements[newRt->stmtCount]);
  st->level = 0;
  st->when.d = convertTime(finalTime);
  st->cmd = "";

  /* Clean up rt so that it may be reused */
  rt->requestNumber = 0;
  rt->requestLine = NULL;
  rt->stmtCount = 0;
  rt->strings = MpCreate();

  /* Insert newRT into the storage list */
  storeRequest(newRt);

  return;
}



int _parentInitStoredListing(Tcl_Interp *interp, void *ctx) {
  Tcl_CreateObjCommand(interp,
                       TCL_TRACE_GET_REQUEST_LISTING_CMD,
                       GetRequestListCmd,
                       NULL,
                       NULL);
  Tcl_CreateObjCommand(interp,
                       TCL_TRACE_GET_REQUEST_INFO_CMD,
                       GetRequestInfoCmd,
                       NULL,
                       NULL);
  return NS_OK;
}



void _initializeStoredRequestHandling(char *hServer) {
  /* Get configuration options */
  if (Ns_ConfigGetInt(TCL_TRACE_PARAMETER_SECTION,
                      STORED_TRACE_SWEEP_PERIOD_KEY,
                      &g_storedTraceSweepPeriod) != NS_TRUE) {
    g_storedTraceSweepPeriod = STORED_TRACE_SWEEP_PERIOD_DEFAULT;
  }

  if (Ns_ConfigGetInt(TCL_TRACE_PARAMETER_SECTION,
                      STORED_TRACE_MAX_AGE_KEY,
                      &g_storedTraceMaxAge) != NS_TRUE) {
    g_storedTraceMaxAge = STORED_TRACE_MAX_AGE_DEFAULT;
  }

  /* Initialize lock and hashtable for long-term request storage */
  Ns_RWLockInit(&g_storedRequestLock);
  Tcl_InitHashTable(&g_storedRequests, TCL_ONE_WORD_KEYS);

  /* Sets up the get request listing command */
  Ns_TclInitInterps(hServer, _parentInitStoredListing, NULL);

  /* Set up cleanup proc that defines the end of a request */
  Ns_RegisterCleanup(CleanupConnection, NULL);

  /* Schedule the sweeper */
  Ns_ScheduleProc(sweepRequests, NULL, 0, g_storedTraceSweepPeriod);
}






/************************************************************************
 Trace Data Traversal

 These procs provide a generic depth-first traversal algorithm for the
 statement call tree. 

 Currently, the traversal process works against the raw trace log.
 While this is reasonably efficient for full traversals, it may make
 it difficult to do some things, such as grabbing a statement and
 finding all of its ancestors.  It might be useful to consider
 transforming the raw entry log into a tree data structure, which
 would be easier to traverse and would simplify some tasks.

 The traversal process defines a concept of statement 'transparency'.
 Transparent statements do not get passed to the processing function
 that is normally executed on each statement.  They do not show up as
 children of any statement in the traversal either -- rather, the
 children of a transparent statement are treated as if they are
 actually the children of the transparent statement's parent (or, more
 precisely, the transparent statement's most recent non-transparent
 ancestor).  For example, if we have this call graph:
 "foo->bar->baz->quux", and bar and baz are both transparent, quux
 will show up as a child of foo during the traversal process.

 In addition to simple 'transparency', commands may be declared
 'cloaking'.  A cloaking command is not only invisible itself, it
 prevents any of its children from being traversed, either.  This
 transparency mode is intended forutility procs written in C, like
 empty_string_p.

 The primary purpose of transparency is to ensure that time spent in
 control-flow procs like 'if' and 'foreach' is charged against the
 time balance of the caller, which seems to make much more sense than
 treating them as seperate entities to be profiled.  In fact, we
 provide a default behavior that makes all TCL built-ins transparent.
 ************************************************************************/



/* Signature of the function that will be called to actually process
   an entry in the statement tree.

   current:  The current statement being processed
   parent:   The statement that called the current statement
             (adjusted for transparency)
   children: The statements invoked by the current statement
             (adjusted for transparency)
   next:     The next statement invoked after the current and all of
             its children are finished.  This statement may be at
             the same stack depth as the current statement, or closer
             to the toplevel.  It may be a transparent. */
typedef void (StatementProcessor)(void *argument,
                                  StatementTrace *current,
                                  int parentCount,
                                  StatementTrace **parent,
                                  int childCount,
                                  StatementTrace **children,
                                  StatementTrace *next,
                                  double totalTime,
                                  double childTime,
                                  int totalChildCount);


/* Signature of the function that will determine if a particular
   command is 'normal', 'transparent', or 'cloaking'.  */
#define TRANSPARENCY_NORMAL 1
#define TRANSPARENCY_TRANSPARENT 2
#define TRANSPARENCY_CLOAKING 3
typedef int (TransparencyTest)(char *name, void *arg);



/* Structure used to bundle up a lot of walk state that we need to pass around */
typedef struct _requestTreeWalk {
  TransparencyTest *transparencyTest;
  void *transparencyArgument;
  StatementProcessor *process;
  void *processArgument;
  StatementTrace *statements;
  int loc;
} RequestTreeWalk;


double walkRequestTreeHelper(RequestTreeWalk *walk,
                             int parentCount,
                             StatementTrace **parent,
                             int *returnChildCount,
                             StatementTrace **returnChildren) {
  double totalTime, childTime = 0.0;
  int i, transparency, firstLoc, subParentCount, subChildCount, childCount = 0;
  StatementTrace **subParent, *st = &(walk->statements[walk->loc]);

  transparency = walk->transparencyTest(st->cmd, walk->transparencyArgument);

  switch (transparency) {
  case TRANSPARENCY_NORMAL:
    /* If the current statement is normal, it inserts itself
       into the parent vector */
    subParent = parent - 1;
    *subParent = st;
    subParentCount = parentCount + 1;
    break;

  case TRANSPARENCY_TRANSPARENT: 
    /* If the current statement is transparent, it simply passes
       on its parent to its children */
    subParent = parent;
    subParentCount = parentCount;
    break;

  case TRANSPARENCY_CLOAKING: 
    /* If a command is 'cloaking', then the entire subtree it
       respresents is invisible, so we just walk the location pointer
       to the next statement that isn't a child of the current. */
    do {
      walk->loc++;
    } while (walk->statements[walk->loc].level > st->level);
    *returnChildren = 0;
    return 0.0;
  }

  /* Store firstLoc and step forward to first potential child */
  firstLoc = walk->loc++;

  /* Loop through children of st, recursively calling the walk on
     them, and accumulating children in a StatementVector. */
  while (walk->statements[walk->loc].level > st->level) {
    childTime += walkRequestTreeHelper(walk,
                                       subParentCount,
                                       subParent,
                                       &subChildCount,
                                       returnChildren + childCount);
    childCount += subChildCount;
  }

  /* If the current statement is transparent, then its children have
     already been appended to returnChildren, and we're done. */
  if (transparency == TRANSPARENCY_TRANSPARENT) {
    *returnChildCount = childCount;
    return childTime;
  }

  totalTime = walk->statements[walk->loc].when.d - st->when.d;

  /* Actually process the current statement */
  walk->process(walk->processArgument,
                st,
                parentCount,
                parent,
                childCount,
                returnChildren,
                &(walk->statements[walk->loc]),
                totalTime,
                childTime,
                walk->loc - firstLoc);

  /* Only one child results, append it to returnChildren */
  *returnChildCount = 1;
  *returnChildren = st;

  return totalTime;
}


void walkRequestTree(RequestTrace *rt,
                     TransparencyTest transparencyTest,
                     void *transArg,
                     StatementProcessor process,
                     void *processArg) {
  RequestTreeWalk walk;
  int childCount;
  StatementTrace *childArray[10000], *parentArray[500];

  /* Prepare walk context data structure */
  walk.transparencyTest = transparencyTest;
  walk.transparencyArgument = transArg;
  walk.process = process;
  walk.processArgument = processArg;
  walk.statements = rt->statements;
  walk.loc = 0;

  /* Run it */
  walkRequestTreeHelper(&walk, 0, &(parentArray[500]), &childCount, childArray);
}
                     





/************************************************************************
 Statistics Computation                                                                        

 These procs employ the generic traversal tools above to gather
 statistics about the commands invoked during a request.  A TCL
 command to retrieve statistics for a particular request is provided.
 ************************************************************************/


#define STATISTICS_TRANSPARENT_KEY "TransparentCommands"
#define STATISTICS_TRANSPARENT_DEFAULT ""

#define STATISTICS_CLOAKING_KEY "CloakedCommands"
#define STATISTICS_CLOAKING_DEFAULT ""

/* 'Transparent' commands are those that don't get profiled themselves.
   Instead, the time spent executing them is counted as belonging to the
   procs from which they are called, and their children are treated as
   the children of procs from which *they* are called. */

static char *g_transparentCommands[] = {
  /* TCL builtins */
  "after", "append", "array", "bgerror", "binary", "break", "catch", "cd",
  "clock", "close", "concat", "continue", "dde", "encoding", "eof", "error",
  "eval", "exec", "exit", "expr", "fblocked", "fconfigure", "fcopy", "file",
  "fileevent", "filename", "flush", "for", "foreach", "format", "gets",
  "glob", "global", "history", "if", "incr", "info", "interp", "join",
  "lappend", "lindex", "linsert", "list", "llength", "load", "lrange",
  "lreplace", "lsearch", "lsort", "memory", "namespace", "open", "package",
  "parray", "pid", "proc", "puts", "pwd", "read", "regexp", "registry",
  "regsub", "rename", "resource", "return", "scan", "seek", "set", "socket",
  "source", "split", "string", "subst", "switch", "tell", "time", "trace",
  "unknown", "unset", "update", "uplevel", "upvar", "variable", "while",

  /* Terminator */
  NULL
};

Tcl_HashTable g_transparentCommandHash;
int g_transparentPatternCount;
char **g_transparentPatterns;


int g_cloakingPatternCount;
char **g_cloakingPatterns;

typedef struct {
  Tcl_Interp *interp;
  int cloakingCount;
  Tcl_Obj *cloaking;
  int transparentCount;
  Tcl_Obj *transparent;
} TransparencyParameters;

/* Standard transparency predicate
 
   This predicate identifies all TCL built-ins listed in the array
   above as transparent, and all arg parser procs generated by ad_proc
   as cloaked. */
int standardTransparency(char *cmd, void *arg) {
  int i;
  Tcl_HashEntry *he;
  TransparencyParameters *tp = (TransparencyParameters *)arg;
  Tcl_Obj *listEntry;

  /* 'Cloaking' overrides 'transparent', so we check that first */
  for (i = 0; i < g_cloakingPatternCount; i++) {
    if (Tcl_StringMatch(cmd, g_cloakingPatterns[i])) {
      return TRANSPARENCY_CLOAKING;
    }
  }
  for (i = 0; i < tp->cloakingCount; i++) {
    Tcl_ListObjIndex(tp->interp, tp->cloaking, i, &listEntry);
    if (listEntry != NULL && strcmp(Tcl_GetString(listEntry), cmd) == 0) {
      return TRANSPARENCY_CLOAKING;
    }
  }

  /* Now check for transparency */
  he = Tcl_FindHashEntry(&g_transparentCommandHash, cmd);
  if (he != NULL) {
    return TRANSPARENCY_TRANSPARENT;
  } 
  for (i = 0; i < g_transparentPatternCount; i++) {
    if (Tcl_StringMatch(cmd, g_transparentPatterns[i])) {
      return TRANSPARENCY_TRANSPARENT;
    }
  }
  for (i = 0; i < tp->transparentCount; i++) {
    Tcl_ListObjIndex(tp->interp, tp->transparent, i, &listEntry);
    if (listEntry != NULL && strcmp(Tcl_GetString(listEntry), cmd) == 0) {
      return TRANSPARENCY_TRANSPARENT;
    }
  }

  /* Default case */
  return TRANSPARENCY_NORMAL;
}


/* Do-nothing transparency test, always returns 'normal' */
int noTransparency(char *ignored1, void *ignored2) {
  return TRANSPARENCY_NORMAL;
}


/* Statistics associated with a particular command are accumulated in
   an instance of this structure. */
typedef struct _statisticsEntry {
  char *cmd;
  int callCount;
  double totalTime;
  double childTime;
  double localTime;
  double averageLocalTime;
  int totalChildCount;
  double averageChildCount;
} StatisticsEntry;



typedef struct {
  double totalDuration;
  double reportedChildDuration;
  double calculatedChildDuration;
} TraversalTraceEntry;


typedef struct {
  RequestTrace *rt;
  TraversalTraceEntry *entries;
} TraversalTraceState;


void TraversalTraceStatementProcessor(void *argument,
                                      StatementTrace *current,
                                      int parentCount,
                                      StatementTrace **parent,
                                      int childCount,
                                      StatementTrace **children,
                                      StatementTrace *next,
                                      double totalDuration,
                                      double childDuration,
                                      int totalChildCount) {
  TraversalTraceState *tts = (TraversalTraceState *)argument;
  TraversalTraceEntry *tte;
  StatementTrace *statements = tts->rt->statements;
  int index, i;

  index = current - statements;
  tte = tts->entries + index;

  tte->totalDuration = totalDuration;
  tte->reportedChildDuration = childDuration;
  tte->calculatedChildDuration = 0.0;
  for (i = 0; i < childCount; i++) {
    tte->calculatedChildDuration +=
      tts->entries[children[i] - statements].totalDuration;
  }
}



int GetRequestTraceCmd(ClientData dummy,
                       Tcl_Interp *interp,
                       int argc,
                       Tcl_Obj *CONST argv[]) {
  RequestTrace *rt;
  int requestNumber, i;
  TraversalTraceState tts;
  register TraversalTraceEntry *tte;
  register StatementTrace *st;
  char entryBuffer[1000];
  Tcl_Obj *result, *resultEntry[5];

  if (argc != 2) {
    Tcl_AppendResult(interp,
                     "wrong # args: should be: \"",
                     Tcl_GetString(argv[0]), " requestNumber\"", NULL);
    return TCL_ERROR;
  }

  if (Tcl_GetIntFromObj(interp, argv[1], &requestNumber) != TCL_OK) {
    Tcl_SetResult(interp, "parameter must be an integer", TCL_STATIC);
    return TCL_ERROR;
  }

  rt = retrieveRequest(requestNumber);

  tts.rt = rt;
  tts.entries =
    (TraversalTraceEntry *)Ns_Calloc(rt->stmtCount,
                                     sizeof(TraversalTraceEntry));
  walkRequestTree(rt,
                  noTransparency, NULL,
                  TraversalTraceStatementProcessor, &tts);

  result = Tcl_NewObj();
  for (i = 0, st = rt->statements, tte = tts.entries;
       i < rt->stmtCount; 
       i++, st++, tte++) {
    /* Build up the entry record :
         cmdName
         stackLevel
         totalDuration
         reportedChildDuration
         calculatedChildDuration
    */
    resultEntry[0] = Tcl_NewStringObj(st->cmd, -1);
    resultEntry[1] = Tcl_NewIntObj(st->level);
    resultEntry[2] = Tcl_NewDoubleObj(tte->totalDuration);
    resultEntry[3] = Tcl_NewDoubleObj(tte->reportedChildDuration);
    resultEntry[4] = Tcl_NewDoubleObj(tte->calculatedChildDuration);
    
    /* Construct Entry list record to result list */
    Tcl_ListObjAppendElement(interp,
                             result,
                             Tcl_NewListObj(5, resultEntry));
  }

  Ns_Free(tts.entries);

  Tcl_SetObjResult(interp, result);
  
  return TCL_OK;
  
}





/* This function is used with walkRequestTree to gather statistics
   about the commands invoked during a request. */
void StatisticsStatementProcessor(void *argument,
                                  StatementTrace *current,
                                  int parentCount,
                                  StatementTrace **parent,
                                  int childCount,
                                  StatementTrace **children,
                                  StatementTrace *next,
                                  double totalDuration,
                                  double childDuration,
                                  int totalChildCount) {
  Tcl_HashTable *statistics = (Tcl_HashTable *)argument;
  int newEntryP, i;
  double localDuration;
  StatisticsEntry *se;
  Tcl_HashEntry *he;
  StatementTrace *st;

  /* If the command is not transparent, create an entry and return totalTime */
  he = Tcl_CreateHashEntry(statistics, current->cmd, &newEntryP);
  if (newEntryP) {
    /* This is freed when we transfer hash contents to array */
    se = (StatisticsEntry *)Ns_Malloc(sizeof(StatisticsEntry));
    se->cmd = current->cmd;
    se->callCount = 0;
    se->totalTime = 0.0;
    se->childTime = 0.0;
    se->localTime = 0.0;
    se->averageLocalTime = 0.0;
    se->totalChildCount = 0;
    se->averageChildCount = 0.0;
    Tcl_SetHashValue(he, (ClientData)se);
  } else {
    se = (StatisticsEntry *)Tcl_GetHashValue(he);
  }

  se->callCount++;
  se->totalTime += totalDuration;
  se->childTime += childDuration;
  se->totalChildCount += totalChildCount;
  se->localTime += (totalDuration - childDuration);
  se->averageLocalTime = se->localTime / se->callCount;
  se->averageChildCount = se->totalChildCount;
  se->averageChildCount /= se->callCount;
}



/* Comparison function used with qsort to sort an array of
   StatisticsEntries by the entry's total execution time */
int TotalCompare(const void *u, const void *v) {
  StatisticsEntry
    *su = (StatisticsEntry *)u,
    *sv = (StatisticsEntry *)v;

  if (su->totalTime < sv->totalTime) {
    return -1;
  }

  if (su->totalTime > sv->totalTime) {
    return 1;
  }

  return 0;
}



/* Comparison function used with qsort to sort an array of
   StatisticsEntries by the entry's local execution time */
int LocalCompare(const void *u, const void *v) {
  StatisticsEntry
    *su = (StatisticsEntry *)u,
    *sv = (StatisticsEntry *)v;

  if (su->localTime < sv->localTime) {
    return 1;
  }

  if (su->localTime > sv->localTime) {
    return -1;
  }

  return 0;
}



/* Comparison function used with qsort to sort an array of
   StatisticsEntries by the entry's 'average' local execution time */
int AverageLocalCompare(const void *u, const void *v) {
  StatisticsEntry
    *su = (StatisticsEntry *)u,
    *sv = (StatisticsEntry *)v;

  if (su->averageLocalTime < sv->averageLocalTime) {
    return -1;
  }

  if (su->averageLocalTime > sv->averageLocalTime) {
    return 1;
  }

  return 0;
}



/* TCL command to fetch a listing of statistics for a request */
int GetRequestStatisticsCmd(ClientData dummy,
                            Tcl_Interp *interp,
                            int argc,
                            Tcl_Obj *CONST argv[]) {
  Tcl_HashTable statisticsHash;
  Tcl_HashSearch hashIter;
  Tcl_HashEntry *he;
  StatisticsEntry *statArray, *se;
  RequestTrace *rt;
  int i, requestNumber, entryCount, entryCapacity;
  TransparencyParameters tp;
  Tcl_Obj *result, *resultEntry[8];

  if (argc < 2 || argc > 4) {
    Tcl_AppendResult(interp,
                     "wrong # args: should be: \"",
                     Tcl_GetString(argv[0])
                     , " requestNumber [transparentList [cloakedList]]\"",
                     NULL);
    return TCL_ERROR;
  }

  if (Tcl_GetIntFromObj(interp, argv[1], &requestNumber) != TCL_OK) {
    Tcl_SetResult(interp, "requestNumber must be an integer", TCL_STATIC);
    return TCL_ERROR;
  }

  tp.interp = interp;

  if (argc > 3) {
    tp.transparent = argv[2];
    if (Tcl_ListObjLength(interp, argv[2], &(tp.transparentCount)) != 
        TCL_OK) {
      Tcl_SetResult(interp, "transparentList must be a list", TCL_STATIC);
      return TCL_ERROR;
    }
  } else {
    tp.transparentCount = 0;
  }             

  if (argc == 4) {
    tp.cloaking = argv[3];
    if (Tcl_ListObjLength(interp, argv[3], &(tp.cloakingCount)) !=
        TCL_OK) {
      Tcl_SetResult(interp, "cloakingList must be a list", TCL_STATIC);
      return TCL_ERROR;
    }
  } else {
    tp.cloakingCount = 0;
  }

  rt = retrieveRequest(requestNumber);

  if (rt == NULL) {
    Tcl_SetResult(interp, "Unknown request number", TCL_STATIC);
    return TCL_ERROR;
  }

  Tcl_InitHashTable(&statisticsHash, TCL_STRING_KEYS);

  /* Walk the tree and build up the table of statistics */
  walkRequestTree(rt,
                  standardTransparency, &tp,
                  StatisticsStatementProcessor,
                  &statisticsHash);
                  
  /* Transfer data from hash to array, for sorting */

  entryCount = 0;
  entryCapacity = 64;

  /* This is freed just before we return from this function */
  statArray =
    (StatisticsEntry *)Ns_Calloc(entryCapacity, sizeof(StatisticsEntry));

  for (he = Tcl_FirstHashEntry(&statisticsHash, &hashIter);
       he != NULL;
       he = Tcl_NextHashEntry(&hashIter)) {

    /* Wouldn't it be nice if we had a way of getting a hashtable's
       entry count, so that we didn't have to do this? */
    if (entryCapacity <= entryCount) {
      entryCapacity <<= 1;
      statArray =
        (StatisticsEntry *)Ns_Realloc(statArray,
                                      entryCapacity * sizeof(StatisticsEntry));
    }
    se = (StatisticsEntry *)Tcl_GetHashValue(he);
    statArray[entryCount++] = *se;
    Ns_Free(se);
  }

  Tcl_DeleteHashTable(&statisticsHash);

  /* Sort the statistics by localTime */
  qsort(statArray, entryCount, sizeof(StatisticsEntry), &LocalCompare);

  /* Build up TCL result */
  result = Tcl_NewObj();
  for (i = 0; i < entryCount; i++) {
    se = &(statArray[i]);

    resultEntry[0] = Tcl_NewStringObj(se->cmd, -1);
    resultEntry[1] = Tcl_NewIntObj(se->callCount);
    resultEntry[2] = Tcl_NewDoubleObj(se->totalTime);
    resultEntry[3] = Tcl_NewDoubleObj(se->childTime);
    resultEntry[4] = Tcl_NewDoubleObj(se->localTime);
    resultEntry[5] = Tcl_NewDoubleObj(se->averageLocalTime);
    resultEntry[6] = Tcl_NewIntObj(se->totalChildCount);
    resultEntry[7] = Tcl_NewDoubleObj(se->averageChildCount);

    Tcl_ListObjAppendElement(interp,
                             result,
                             Tcl_NewListObj(8, resultEntry));
  }

  Ns_Free(statArray);

  Tcl_SetObjResult(interp, result);
  
  return TCL_OK;
}



int _parentInitStatistics(Tcl_Interp *interp, void *ctx) {
  Tcl_CreateObjCommand(interp,
                       TCL_TRACE_GET_REQUEST_STATISTICS_CMD,
                       GetRequestStatisticsCmd,
                       NULL,
                       NULL);
  Tcl_CreateObjCommand(interp,
                       TCL_TRACE_GET_REQUEST_TRACE_CMD,
                       GetRequestTraceCmd,
                       NULL,
                       NULL);
  return NS_OK;
}


void _initializeStatistics(char *hServer) {
  char **cmd, *cfgString, *cfgEntry, rootCmd[1000];
  int i, newFlag;
  
  rootCmd[0] = rootCmd[1] = ':';

  Tcl_InitHashTable(&g_transparentCommandHash, TCL_STRING_KEYS);
  
  for (cmd = g_transparentCommands; *cmd != NULL; cmd++) {
    Tcl_CreateHashEntry(&g_transparentCommandHash, *cmd, &newFlag);
    /* This adds the command to the hash a second time, with "::"
       prepended, for those situations where the commands are
       explicitly invoked from the root namespace. */
    strcpy(&(rootCmd[2]), *cmd);
    Tcl_CreateHashEntry(&g_transparentCommandHash, rootCmd, &newFlag);
  }

  g_transparentPatternCount = 0;
  cfgString = Ns_ConfigGetValue(TCL_TRACE_PARAMETER_SECTION,
                                STATISTICS_TRANSPARENT_KEY);
  if (cfgString != NULL) {
    /* Never freed -- lives until server shutdown */
    cfgString = Ns_StrDup(cfgString);
    for (cfgEntry = strtok(cfgString, ",\n");
         cfgEntry != NULL;
         cfgEntry = strtok(NULL, ",\n")) {
      if (g_transparentPatternCount == 0) {
        g_transparentPatterns = (char **)Ns_Malloc(sizeof(char *));
      } else {
        g_transparentPatterns =
          (char **)Ns_Realloc(g_transparentPatterns,
                              (g_transparentPatternCount+1) * sizeof(char *));
      }
      g_transparentPatterns[g_transparentPatternCount++] = cfgEntry;
    }
  }

  cfgString = Ns_ConfigGetValue(TCL_TRACE_PARAMETER_SECTION,
                                STATISTICS_CLOAKING_KEY);
  g_cloakingPatternCount = 0;
  if (cfgString != NULL) {
    /* Never freed -- lives until server shutdown */
    cfgString = Ns_StrDup(cfgString);
    for (cfgEntry = strtok(cfgString, ",\n");
         cfgEntry != NULL;
         cfgEntry = strtok(NULL, ",\n")) {
      if (g_cloakingPatternCount == 0) {
        g_cloakingPatterns = (char **)Ns_Malloc(sizeof(char *));
      } else {
        g_cloakingPatterns =
          (char **)Ns_Realloc(g_cloakingPatterns,
                              (g_cloakingPatternCount+1) * sizeof(char *));
      }
      g_cloakingPatterns[g_cloakingPatternCount++] = cfgEntry;
    }
  }

  Ns_TclInitInterps(hServer, _parentInitStatistics, NULL);
}





/* Initialization function invoked by AOLserver when the module is loaded */
NS_EXPORT int
Ns_ModuleInit(char *hServer, char *hModule) {
  Ns_Log(Notice, "TCLPROF: initializing on %s", hServer);

  _initializeMemoryPool();
  _initializeTrace(hServer);
  _initializeStoredRequestHandling(hServer);
  _initializeStatistics(hServer);

  return NS_OK;
}
